;-===============================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-	GNU GPL, restricted to non-commercial useage, read
;-	LICENSE file for additional copyright information.
;-
;-  file  	: cpuid.asm
;-  package	: "cpuid.tar.gz" with
;-		:	cpuid, cpuid.asm, cpucaps.inc, cpuid.defs.inc
;-		;	cpucaps.txt.gz, LICENSE, mkg, debug.inc
;- 		: latest version(s) available at
;-		:	http://www.snafu.de/~phpr/cpuid.tar.bz2
;-  created	: 10-mar-2000
;-  modified	: 20-mai-2000
;-  version	: 0.23
%define VER '0.24'
;-  assembler	: nasm 0.98
;-  description	: get specified cpu feature flag(s) or full cpu specification.
;-		: ececute <mkg> for glossary "cpucaps.gls" from "cpucaps.txt.gz"
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: earliest processor type recognized is (should be) i386.
;-		: signal handler (re "ssef") somewhat obscure (uncertain?).
;-  source	1) AMD Processor Recognition Application Note #20734, 11/99
;-		; http://www.amd.com/products/cpg/athlon/techdocs/index.html
;-		2) Intel Processor Identification and the CPUID Instruction
;-		: AP-485 app. note, #241618012 (1/99); http://www.intel.com
;-		3) Intel Architecture Developers Manual, 1999, Vol.1-3 (3),
;-		: ref. #243190, #243191, #243192; http://www.intel.com
;-		4) opcodes list 1/1997 by 'Potemkin's Hackers Group'
;-		: mail:root@aleste.misa.ac.ru (no web/ftp address, no response).
;-		5) Linux 2.2.14 sources, "arch/i386/kernel/head.S"
;-  requires	: cpuid.defs.inc, cpucaps.inc, signal_h.inc
;-		; (optionally, to debugging: debug.inc, cpuid.test.inc)
;-		: system.inc, etc, from asmutils 0.08, http://linuxassembly.org
;-		: i386+ IBM-PC style hardware & Linux 2.2+ (2.2.14,i586 tested)
;-===============================================================================;
;-
;-

;-
;- comments:
;-	i:	input data
;-	o:	output data
;-	c:	changed items
;-	p:	preserved items
;-	<[val]>	effective address
;-	A	AMD specific
;-	I	Intel specific
;-	O	other manufacturers
;-	X	vendor/model independent
;-	-/-	item doesn't apply
;-

%include "system.inc"
%include "cpuid.defs.inc"

%undef saflags
%define saflags SA_RESETHAND	;(for asm2mkh/modified nasm)
%define saflags SA_ONESHOT	;(for asm2mkh/modified nasm)
%define saflags SA_SIGINFO	;(for asm2mkh/modified nasm)

;- system constants should be found automatically,
;- uncomment if script <create-file-mkh> fails:
;-%include "signal_h.inc"
;-

    CODESEG
Z:
;- Z
;-	beginning of <.text> segment

;-	<cpucaps.inc> query and response strings
%include "cpucaps.inc"

;-	outcommenting <a> alignment would save about 300 bytes
%define a align 8

;-===============================================================================;
;-
;- debugging:
%ifdef __DBG
;-	extra code assembled if _DBG defined, else just the labels & ret.
;- to enabling define prior to <make>:
;-	export ASFLAGS=-D__DBG;make
;- include routine & dummy data to test,
;- includes nothing if not both, __TEST and __DBG, are defined:
;-	export ASFLAGS="-D__DBG -D__TEST";make
;- or hard-code in "cpuid.defs.inc"
;-
;-	<other/hp/debug.inc> debugging stuff, macroes and data
;-	macroes PFSYM, PRNUM, PRSTG, PRTXT, etc - re source gls.
%include "debug.inc"
;-	<cpuid.test.inc> debugging with alien processor data
%include "cpuid.test.inc"

dbg:
	mdump
	perrm
	perrs
	perrn
	dbmrg
	dbcons	; ends up in _DATA
    section .text
	pbits	; flagbits/binary
	psbits	; eflags symbols
%endif

;-
;- cpu_id
;-	re-definable substitute to <cpuid> instruction, for testing
;-	defaults to the real instrunction if __TEST_I not defined 
;-	(from cpuid.test.inc, included if __DGB and __TEST def'd).
;-	re also <chk_cpu>, below
%ifndef __TEST_I
%define cpu_id  cpuid
%define chk_cpu check_cpu
%endif

;-===============================================================================;

	align 32
;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	jump tables & associated data								;
;-
;- jmptab
;-	main jumptable
;  (long and memory inefficient but, simple...)
jmptab:
	dd unkn,help,lhel,vers,lver,strg,obin,cnum	; ..07
	dd ooct,lnam,sbl ,ohex,rstg,rbin,rcnm,roct	; ..15
	dd rlnm,rbl ,rohx,qust,unkn,amd ,intl,cyrx	; ..23
	dd cntr,umc ,unkn,unkn,noid,unkn,i386,i486	; ..31
	dd i586,i686,i786,a486,k5  ,k6  ,k6_2,k6_3	; ..39
	dd atln,i4dx,i4sx,i4d2,i4sl,i4x2,i4e2,i4x4	; ..47
	dd p586,i5mx,i6pr,i6p2,i6ce,i6p3,u386,u486	; ..55
	dd u586,u686,u786,vend,type,vstg,name,cpid	; ..63
	dd feat,xtyp,regs,cach,oem ,ovdv,dual,ityp	; ..71
	dd ssem,ssef,psn ,fpu ,vme ,de  ,pse ,tsc 	; ..79
	dd msr ,pae ,mce ,cx8 ,apic,unkn,sep ,mtrr	; ..87
	dd pge ,mca ,cmov,pat ,ps36,pn  ,unkn,unkn	; ..95
	dd unkn,xmmx,mmx ,fxsr,sse ,unkn,unkn,unkn	; ..03
	dd l64m,x3dn,a3dn,simd,x1da,x1de,x1ia,x1ie	; ..11
	dd t1da,t1de,t1ia,t1ie,l1ds,l1da,l1cl,l1cz	; ..19
	dd l1is,l1ia,l1il,l1iz,x2da,x2de,x2ia,x2ie	; ..27
	dd t2da,t2de,t2ia,t2ie,l2us,l2ua,l2ul,l2uz	; ..35
	dd it32,it02,dt4k,dt4m,ic8 ,ic16,dc8 ,dc16	; ..43
	dd nl2 ,uc1 ,uc2 ,uc5 ,uc1m,uc2m,noti	; ..51
; ovf:
    dd noti
;- typet
;-	processor vendors
typet:
    dd utype,amdy,intly,cyrxy,cntry,umcy,utype,utype
    dd utype,utype,utype,utype,utype,utype,utype,utype
;- amdyt
;-	processor models AMD
amdyt:
    dd utype,utype,utype,utype,a486y,ak6y,atlony,utype
    dd utype,utype,utype,utype,utype,utype,utype,utype
;- intlyt
;-	Intel processor families
intlyt:
    dd utype,utype,utype,utype,i4ty,i5ty,i6ty,utype
    dd utype,utype,utype,utype,utype,utype,utype,utype
;- aspec, asfbm, axfbm, isfbm
;-	valid features bits masks
;  amd 10987654321098765432109876543210
aspec:
    dd 00000000000000000011001000000000b
asfbm:
    dd 00000001100000111111101111111111b
;md-x  3	  2	    1	      0
axfbm:
    dd 11000001110000111111101111111111b
;intel 10987654321098765432109876543210
isfbm:
    dd 00000011100001111111101111111111b
;- avail
;-	control sequence for '-s -?'
avail:	; +ve query index, -ve:loop count, 0:end ;
    db (CAP.VENDOR-QUERY.BOT)/4,(CAP.TYPE-QUERY.BOT)/4
    db (CAP.XTYP-QUERY.BOT)/4,(CAP.FEAT-QUERY.BOT)/4
    db (CAP.SSE_MMX-QUERY.BOT)/4
acach:
    db (L1.BOT-QUERY.BOT)/4,(L1.BOT-L2.UC2M)/4,0
;- icache
;-	intel cache config. codes
icache:	; keep in sync with querystrings ;
    db 1,2,3,4,6,8,10,12,64,65,66,67,68,69
.e:
;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	cpu recognition & data acquisition							;

;-
;- check_cpu
;-	check whether cpuid functional, subroutine
;-	called via %define-d <chk_cpu>
;- i:	ebx	bitmask
;- o	eax	flag (0:no 1:success)
;-	eflags	Z/NZ, S/NS
;- c:	eax,ebx,edx,eflags
    a
check_cpu:
    mov [byte esi-vref+signature],eax			; pre-set family
    pushfd						; save EFLAGS to stack.
    pop eax
						; store EFLAGS in EAX.
    mov edx,eax						; save in EBX for testing later.
    xor eax,ebx						; toggle flag bit
    push eax						; copy "changed" value to stack.
    popfd						; save "changed" EAX to EFLAGS.
    pushfd
    pop eax
    push edx
    popfd						; restore flags
    xor eax,edx						; whether bit changeable.
    jz .r						;?; ret Z if not
    xor eax,eax
    inc eax						; else NZ+NS+PO
.r:
    ret

;-
;- cpu_type
;-	determine cpu type and cpuid modes
;- i:	-/-
;- o:	eflags	S:no cpuid, Z:unfeatured cpuid, NZ&NS:featured cpuid
;-	eax	flag (-ve:no cpuid, 0,+ve:cpuid functional)
;-	ebx,ecx,edx vendor string if cpuid succeeded
;- c:	eax,ebx,ecx,edx,edi
    a
cpu_type:
    xor eax,eax
    dec eax
    mov [byte esi-vref+clevel],eax			; prepare -1 for 'no cpuid'
    mov ebx,1<<14					; AC bit
    mov eax,0x0300					; type 386: synthetic "family" 3
    call chk_cpu
    jz .r						;?; no cpuid: ret -ve
    mov eax,0x0400					; type 486 w.o. cpuid
    mov ebx,1<<21					; ID bit
    call chk_cpu
    jz .r						;?; -ve if 486 w.o. cpuid
    lea edi,[byte esi-vref+clevel]

    xor eax,eax						; #0: eax:=max. std queries, b,d,c vendor
    call rcpid
    mov [edi+c_c],edx					; re-arrange for vendor string
    mov [edi+c_d],ecx
    inc eax
.r:
    dec eax
    ret
;-
;- rcpid
;-	call cpuid and store ret. regs
;- i:	eax cpuid level
;-	edi ptr to reg. storage
    a
rcpid:
    xor ebx,ebx
    mov ecx,ebx
    mov edx,ebx
    cpu_id
    mov [edi+c_a],eax
    mov [edi+c_b],ebx
    mov [edi+c_c],ecx
    mov [edi+c_d],edx
    ret
;-
;- i_cpu
;-	fetch intel additional levels and cache/TLB descriptors
;- i:	esi	vari ptr
;- o:	-/-
;- c:	all but esi
;-	level 2+ at <xfeat>+, cache levels at <intel20>+
    a
i_cpu:
    mov eax,[byte esi-vref+clevel]
    and eax,byte 15					; truncate to safe max.
    jz .r
    movzx ebx,al					; levels
    mov [byte esi-vref+clnum],eax			; initiate count & command value
    add ebx,ebx
    lea edi,[byte esi-vref+xfeat+ebx*8]			; top level
.l:
    mov eax,[byte esi-vref+clnum]
    call rcpid						; store this level-s regs
    sub edi,byte cb
    dec byte[byte esi-vref+clnum]
    jns .l						; loop through all levels
    mov eax,[byte esi-vref+clevel]
    cmp al,2
    js .r
    lea edi,[byte esi-vref+intel_20]			; low 2nd level
.i:
    mov eax,2						; level command remains same, return byte 0 tells levels to follow
    call rcpid
    add edi,byte cb
    cmp al,1
    jg .i
.r:
    ret

;-
;- detect_base
;-	collect cpu info, check for co-proc. if type is 386
;- i:	-/-

;- o:	eax	flag (-ve:no 1:cpuid functional)
;	 al	cpu type
    a
detect_base:
    call cpu_type
    mov [byte esi-vref+ibase],eax
    jle .r						;?; no cpuid (S) or no xt'd features (Z)
    xor eax,eax
    inc eax						;#1: Check for processor features
    lea edi,[byte esi-vref+signature]
    call rcpid
    mov eax,0x80000000					;#8x0: eax:=0x80000000+max. extended cpuid queries
    cpu_id
    and eax,0x8000000f					; max 15 loops assumed, just to get safe
    mov [byte esi-vref+xlevel],eax
;; Check which extended functions can be called		;
    jns i_cpu						;I: +ve: check for intel standard xtns
    mov edi,eax
    test eax,eax
    jz .c
    mov [byte esi-vref+clnum],eax			; store std levels counter
;;6: L2 :5: L1 :4,3,2:A:name string :1:xt'd features	;
    movzx eax,al
    add al,al
    lea edi,[byte esi-vref+xfeat+eax*8-16]
.n:
    mov eax,[byte esi-vref+clnum]
    call rcpid
    sub edi,byte cb
    mov eax,[byte esi-vref+clnum]
    dec al
    mov [byte esi-vref+clnum],eax
    jg .n						; don't level 80..00
    xor eax,eax
    inc eax
    inc eax						; Set "Has cpuid" flag to true
.c:
    dec eax
.r:
    ret

;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	query response										;
;-
;- getcaps
;-	main query response subroutine
;- i:	ecx	feature no. (re cpucaps.inc)
;-	esi	ptr to variables buffer, <vref>
;-		destn routines further receive:
;-			eax:=ecx
;-			ebx:=0
;-			ebp:=esp (don't modify!)
;-			edi:=[VNAME.NONE]
;- o:	eax	feature present(num) or not(0)
;-	ecx	ptr to id-string
;-	edx	long stg len, if appl.
;-	ebp	<[tos-4]>
;-	edi	ptr to long string, if available - <oflagb>, re defn's above
;-	eflags	Z: no further action (i.e. no output) after return
;- c:	all but esi (esi must be preserved/restored before exit)
    a
getcaps:
    mov eax,ecx						; index
    xor ebx,ebx						; returnflags
;	PRSTG ':'
;	PRNUM ecx,1
    test iflagb,r_qst
    jnz .x						;?; previous was '-?'
.q:
    lea ebp,[esp]					; have ebp in a defined state, in case of signal...
    lea edi,[VNAME.NONE]				; ref. addr & "No Vendor ID"
    jmp dword[jmptab+ecx*4]
.x:							; ret numeric index of query
    cmp ecx,(OPTIONS.QUST-QUERY.BOT)/4
    jz .q						;?; 2nd '-?'
; :HIER: hab den zweck vergessen... :HIER:
;    and iflagb,~r_stg					; no string return
    test esi,esi
    ret
    
;; query response routines				;

    a
cpid:
    mov eax,[byte esi-vref+clevel]			; ret packed no. of cpuid levels
    mov ah,[byte esi-vref+xlevel]			; merge amd extended levels
    lea ecx,[edi-VNAME.NONE+CAP.CPUID]
    test ax,ax
    jnz noid.r
noid:
    lea ecx,[edi-VNAME.NONE+VENDOR.NONE]		; 'noid'
    lea edi,[edi-VNAME.NONE+noidt]
.n:
    mov eax,ebx						; returncode 0
    mov oflagb,o_ascil|o_lstg				; asciz long, available
    inc ebx						; NZ
.r:
    ret
    a
noti:
    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]			; 'unkn'
    lea edi,[edi-VNAME.NONE+quert]
    jr noid.n
    a
lhel:
help:
    mov bx,(o_ascil|o_ascis)<<8|r_lstg			; l+s is asciz, query forced to long string
    jr vers.h
    a
lver:
    test iflagb,~r_lstg
    jnz vers
    or bl,r_lstg					; default to long stg
vers:							; id is fixed len
    or bl,iflagb
    and bl,~r_digs					; suppress numeric query
    push byte(helpt.e-helpt)
    pop edx						; long version length
.h:
    or bx,(o_lstg|o_quit)<<8|r_stg			; long string available, force string type query
    mov ioflagw,bx
    lea ecx,[edi-VNAME.NONE+helpt]		; help/version text
    mov edi,ecx						; long=short
    inc ebx						; NZ, enable output
    ret

; set output options
    a
sbl:
    mov bl,r_bl
    jr sflag
lnam:							; -t: bit 1: verbose strings
    mov bl,r_lstg
    jr sflag
strg:							; -s: bit 0: identifier strings
    mov bl,r_stg
    jr sflag
cnum:							; -u: bit 7: numeric
    mov bl,r_num
    jr sflag
obin:							; -b: bit 4: binary
    mov bl,r_bin
    jr sflag
ooct:							; -o: bit 5: octal
    mov bl,r_oct
    jr sflag
ohex:							; -x: bit 6: if numeric ret sedecimal 0x...
    mov bl,r_hex
sflag:							; accumulate output flags
    or iflagb,bl
    xor ebx,ebx						; Z, continue, no output
    ret
    a
rbl:
    mov bl,~r_bl
    jr rflag
rstg:
    mov bl,~r_stg					; clr bit 0
    jr rflag
rcnm:
    mov bl,~r_num					; clr bit 7
    jr rflag
rlnm:
    mov bl,~r_lstg					; clr bit 1
    jr rflag
rbin:
    mov bl,~r_bin					; clr bit 5
    jr rflag
roct:
    mov bl,~r_oct					; clr bit 5
    jr rflag
rohx:
    mov bl,~r_hex					; clr bit 6
rflag:
    and iflagb,bl 					; clr option
    xor ebx,ebx						; Z, continue, no output
    ret

; check vendor, ret flag or num or id
    a
amd:
    push byte(VENDOR.AMD-VENDOR)/4
.v:	; eax:=num *edi=vendorname(12) edx=long len(12) ecx=*id(4)
    or oflagb,o_flag|o_lstg				; result is flag, strings provided
    push ioflagw
    call vendb
    pop ioflagw						; rst output modifiers
    pop ebx
    cmp eax,ebx
    jz .n
    xor eax,eax
    lea ecx,[VENDOR.UNKN]				; id
    lea edi,[VNAME.UNKN]				; long
.n:
    test ecx,ecx					; NZ
    ret
    a
intl:
    push byte(VENDOR.INTEL-VENDOR)/4
    jr amd.v
    a
cyrx:
    push byte(VENDOR.CYRIX-VENDOR)/4
    jr amd.v
    a
cntr:
    push byte(VENDOR.CTAUR-VENDOR)/4
    jr amd.v
    a
umc:
    push byte(VENDOR.UMC-VENDOR)/4
    jr amd.v

; check type, ret flag

; intel
    a
i786:	inc ebx ; 4
i686:	inc ebx ; 3
i586:	inc ebx ; 2
i486:	inc ebx ; 1
i386:		; 0
    push byte(VENDOR.INTEL-VENDOR.BOT)/4
    push byte(CPU.I386-CPU.BOT)/4
    jr ia4

; any vendor
    a
u786:	inc ebx	; 4
u686:	inc ebx	; 3
u586:	inc ebx	; 2
u486:	inc ebx	; 1
u386:		; 0
    mov eax,[byte esi-vref+signature]
    movzx eax,ah
    and eax,byte 15
    sub eax,byte 3
    cmp ebx,eax
    jnz ia4.u
    lea ecx,[edi-VNAME.NONE+CPU.U386+ebx*4]
    test ecx,ecx					; NZ
    ret
    a
; AMD 486
a486:
    push byte(CPU.A486-CPU.BOT)/4
    push byte(VENDOR.AMD-VENDOR.BOT)/4
ia4:
    call vendb
    pop edx
    pop ebp
;	PRNUM eax
    cmp eax,edx
    jnz .u
    call type
    cmp eax,ebp
.u:
    jnz near unkn.s
    test ecx,ecx
    ret
;
    a
k5:	dec ebx
k6:	dec ebx
k6_2:	dec ebx
k6_3:	dec ebx
atln:	dec ebx
i4dx:	dec ebx
i4sx:	dec ebx
i4d2:	dec ebx
i4sl:	dec ebx
i4x2:	dec ebx
i4e2:	dec ebx
i4x4:	dec ebx
p586:	dec ebx
i5mx:	dec ebx
i6pr:	dec ebx
i6p2:	dec ebx
i6ce:	dec ebx
i6p3:
    add ebx,byte(CPU.Pentium_III-CPU.BOT)/4		; requested type no.
; ret "u..." if no vendor determinable ;
    push ebx
    push ecx
    push ioflagw					; protect ioflags
    call type
    pop ioflagw
    or oflagb,o_flag					; enable flag output
    pop edx						; index
    pop ebx						; type no.
;	PRNUM eax
;	PRNUM ebx
    cmp eax,ebx
    jnz ia4.u
    test esp,esp					; NZ
    ret

; -- ; for Intel multi-proc/proc-type in bits 12/13 re "oem", "dual", etc..
    a
type:
    mov ebx,[byte esi-vref+cputype]			; stored ptr to type-id
    test ebx,ebx
    jz .f						;?; 1st request
    lea edi,[CPU.BOT]					; base of type id-strings
    mov eax,ebx
    mov ecx,ebx						; id-string
    sub eax,edi
    mov edx,[byte esi-vref+signature]
;	PRNUM edx
    mov oflagb,o_dstg
    test iflagb,r_digs	;;;
    jnz .s		;;;
;    mov edi,[byte esi-vref+cpuname]			; ptr to long name
    test edi,edi
    jle .s						;?; no long string (serno)
    mov edi,[byte esi-vref+cpuname]			; ptr to long name
;;    push byte 12
;;    pop edx
    and oflagb,~o_dstg
    or oflagb,o_lstg|o_ascil
.s:
    sar eax,2						; type no.
;	PRSTG' tn:'
;	PRNUM eax,1
;	PRTXT edi,12
;	PRSTG':*:'
    test ecx,ecx
    ret
    a
.f:
    push ioflagw
    call vendb
    pop ioflagw						; clear output flags from <vend>
    and eax,byte m_vendm				; vendor number, get safe
    mov ebx,[byte esi-vref+signature]
    mov cl,bh						; ecx:=family
    and ecx,byte 15					; in bits 8..11
    shr bl,4						; discard stepping, leave model
    movzx ebx,bl					; ebx:=model
    movzx eax,al					; eax:=vendor
    lea edi,[CPU.BOT]					; edi:=base of type id-strings
;	PRSTG' tj:'
;	PRNUM eax,1	; vendor
;	PRNUM ebx,1	; model
;	PRNUM ecx,1	; family
    jmp [typet+eax*4]					; select vendor
; - types -
    a
intly:
;	PRSTG'I'
    mov edx,[byte esi-vref+signature]			; signature for long stg request
    or oflagb,o_dstg|o_lstg
    jmp [intlyt+ecx*4]					; select family
    a
amdy:
;	PRSTG'A'
    lea edx,[byte esi-vref+nstr]			; long is processor namestring
    or oflagb,o_lstg|o_ascil
    jmp [amdyt+ecx*4]					; select family
    a
atlony:
;	PRSTG'O'
    lea ecx,[edi+CPU.ATHLON-CPU.BOT]			; "atln"
    jr a486y.a
; - intel 4xx -
    a
i4ty:
;	PRSTG'I4:'
;	PRNUM ebx,1
    lea ecx,[edi+CPU.I486DX-CPU.BOT+ebx*4]		; i4dx,i4sx,i4d2,i4sl,i4x2,i4e2,i4x4,u486
    sub bl,7
    js a486y.r
    lea ecx,[edi+CPU.I486-CPU.BOT]			;7+ "u486"/unknown
    jr utype.j
; - intel 5xx -
    a
i5ty:
;	PRSTG'I5:'
    dec bl
    js .u
    lea ecx,[edi+CPU.Pentium-CPU.BOT]			; "i586"
    sub bl,3
    js a486y.r
    lea ecx,[edi+CPU.Pentium_MMX-CPU.BOT]		; "i5mx"
    jz a486y.r
.u:
    lea ecx,[edi+CPU.U586-CPU.BOT]			; "u586"/unkn
    jr unkny
; - unknown/undeterminate - eax vendor, ebx model, ecx family
    a
cyrxy:		; CYRIX (National) types yet unknown	:HIER:
cntry:		; CENTAUR (?) types yet unknown		:HIER:
umcy:		; UMC (?) types yet unknown		:HIER:

;<; :unknown: & :typet::amdyt::amdyt::intlyt: ;>;
utype:
;	PRSTG'u'
    mov ecx,[byte esi-vref+signature+1]
    and ecx,byte 15
    test bl,bl
    jz .u
    test byte[byte esi-vref+clevel+1],-1
    js .u
    lea ecx,[edi+CPU.U386-CPU.BOT-12+ecx*4]		; "u{3,4,5,6,7}86"
.j:
    sub bl,8
    js unkny
.u:
    lea ecx,[edi+CPU.UNKN-CPU.BOT]			; "unkn"
    jr unkny
; - amd -
    a
ak6y:
    lea ecx,[edi+CPU.K5-CPU.BOT]			;0..3 "k5"
    sub bl,4
    js a486y.a
    lea ecx,[edi+CPU.K6-CPU.BOT]			;4..7 "k6"
    sub bl,4
    js a486y.a
    lea ecx,[edi+CPU.K6_2-CPU.BOT]			;8 "k6-2"
    dec bl
    js a486y.a
    lea ecx,[edi+CPU.K6_3-CPU.BOT]			;9+ "k6-3"
    jr a486y.a
    a
a486y:
    lea ecx,[edi+CPU.A486-CPU.BOT]			; "a486"
.a:
    mov [byte esi-vref+cpuname],edx
.r:
    mov [byte esi-vref+cputype],ecx
    mov eax,ecx
    sub eax,edi
    mov edi,edx
    shr eax,2						; type is no. in id-string list
    test ecx,ecx
    ret
; - intel 6xx - 0:- 1+3:p 4:- 5:p2, 6:c 7:p3 8+:-
    a
i6ty:
    lea ecx,[edi+CPU.Celeron-CPU.BOT]			;6 "i6ce"
    cmp bl,6
    jz a486y.r
    shr bl,1
    jnc .u
    lea ecx,[edi+CPU.Pentium_Pro-CPU.BOT]		;0  ->	1 "i6pr"
    jz a486y.r
    lea ecx,[edi+CPU.Pentium_II-CPU.BOT]		;1,2->	3,5 "i6p2" P3 M3, M5
    sub bl,2
    jle a486y.r
    lea ecx,[edi+CPU.Pentium_III-CPU.BOT]		;3  ->	7 "i6p3"
    dec bl
    jz a486y.r
.u:
    lea ecx,[edi+CPU.U686-CPU.BOT]			; "u686"
unkny:							; synthetic string from signature for unknown type
; - if unknown ret type/family/model numeric & long -
    push ecx
    mov eax,[byte esi-vref+signature]			; numeric signature
;	PRNUM eax
    call n2s
    and oflagb,~o_ascis
    pop ecx
    test ecx,ecx
    ret

; - extra type information - intel feature bits 11&12 -
    a
ityp:	inc ebx	;4
dual:	inc ebx	;3
ovdv:	inc ebx	;2
oem:	inc ebx	;1
    or oflagb,o_flag					; result is flag
xtyp:
; check & exclusion not necessary (?) :OK:
;    xor eax,eax
;    call vamd						; doesn't apply to AMD
;    jz near unkn					;?; AMD
    push byte 00110000b
    pop eax
    and al,[byte esi-vref+features+1]
    shr al,4
    lea ecx,[edi-VNAME.NONE+CAP.OEM+eax*4]
    inc al						; ret non-zero
    test bl,bl
    jz .r						;?; query was "xtyp"
    cmp al,bl
    setne bl
    dec bl
    and al,bl
.r:
    test ecx,ecx
    ret

; - version string -
    a
vstg:
    test iflagb,r_stg|r_lstg
    jz vend						;?; numeric request, else
    or iflagb,r_lstg					; force long type string request
vendb:
    xor ebx,ebx
vend:
; i:
; ebx	0(!)
; o:
; eax	vendor code
; ecx	ptr to 4 chars vendor id
; edx	vendor name length (12)
; edi	ptr to 12 chars vendor name
; esi	variptr <vref> from cpuid
    push byte 12
    pop ecx						; len
    mov edx,[byte esi-vref+vendi]			; type set flag
    test edx,edx
    jnz .s						; type already found
    lea edx,[byte ebx+VENDOR.NONE-VENDOR]
    test dword[byte esi-vref+clevel+c_b],-1		; check ebx from level 0
    jz .n						; no vendor string received
    cld							; cmp increment
    lea edi,[edi+VNAME.AMD-VNAME.NONE]			; string "authentic...
    lea edx,[byte ebx+VENDOR.AMD-VENDOR]
    call vcmp
    jz .v
    lea edi,[edi+VNAME.INTEL-VNAME.AMD]
    lea edx,[byte ebx+VENDOR.INTEL-VENDOR]
    call vcmp
    jz .v
    lea edi,[edi+VNAME.CYRIX-VNAME.INTEL]
    lea edx,[byte ebx+VENDOR.CYRIX-VENDOR]
    call vcmp
    jz .v
    lea edi,[edi+VNAME.CTAUR-VNAME.CYRIX]
    lea edx,[byte ebx+VENDOR.CTAUR-VENDOR]
    call vcmp
    jz .v
    lea edi,[edi+VNAME.UMC-VNAME.CTAUR]
    lea edx,[byte ebx+VENDOR.UMC-VENDOR]
    call vcmp
    jz .v
    lea edx,[byte ebx+VENDOR.UNKN-VENDOR]
.v:
    lea edi,[byte esi-vref+idstr]			;<;edi:=realname, vendor from cpuid
.n:							;<;no vendorstring received
    not edx						; non-zero flag
    mov [esi-vref+vendi],edx				; store found values
    mov [esi-vref+vendn],edi				; for repeated calls
.s:
    mov edi,[byte esi-vref+vendn]
    not edx						; rst from flag
    mov eax,edx						; disp
    shr eax,2						; eax:=vendor code
    and al,m_vendm					; mask to max <m_vendm>=15
    mov bl,v_vend
    or bl,al
    mov [byte esi-vref+o_vend],bl			; ..stored
    cmp al,(VNAME.UNKN-VNAME.BOT)/12
    jb .i
    xor eax,eax						; not found
.i:
    xchg ecx,edx					; ecx:=disp to vendor-id ,edx:=long len
    lea ecx,[ecx+VENDOR]				; ecx:=ptr to vendor-id
    or oflagb,o_lstg					; long string available
    ret
; --
    a
name:	;:HIER: intl,AMD :OK:
    call type						; type code & name
    push eax
    push ebx
    push ecx
    push ioflagw
    call psn						; long is serial no.
    test eax,eax
    pop ax
    jz .n
;    or ah,o_ascil|o_lstg				; make long asciz available
    and ah,~(o_dstg|o_cstg)				; clr string type numbers
;    mov ioflagw,ax
    mov edx,edi
.n:
    or ah,o_ascil|o_lstg				; make long asciz available
    mov ioflagw,ax
    pop ecx
    pop ebx
    pop eax
    test esp,esp
    ret
;
;; - feature bits, special -							;
    a
ssef:
    call sse
    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]			; "unkn"
    test eax,eax
%ifndef __ssef
    jz .s
    lea ebx,[.s]		; (though I really hate silly 'tricks', predefining stacktop and ret-address...)
%else				; :TEST: with synthetic serno
    lea ebx,[.s]
%endif
    call isegf						; prepare "unkn" return on segfault ("illegal instr")
.o:
    orps xmm0,xmm0					;A: trial & error, only: segfault or functional...
    lea ecx,[edi-VNAME.NONE+CAP.SSE_FP]
.s:
    test ecx,ecx
    ret
;
    a
psn:							;I: serial number as long asciz string
    call pn
    and oflagb,~o_flag					; result is not a flag
%ifndef __ssef
    test eax,eax
    jnz .s
.r:
    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]			; "unkn"
    test ecx,ecx
    ret
    a
.s:
%else		; :TEST: with synthetic serno
    jr .s
    a
.r:
    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]			; "unkn"
    test ecx,ecx
    ret
.s:
    mov dword[serno3],0x0a0b0c0d
    mov dword[serno2],0x00edda00
%endif
    mov eax,[byte esi-vref+serno3]			; lo dword
    push ioflagw
    and iflagb,~r_num					; suppress leading zeroes suppression
    call retnum						; start conversion
    add ecx,byte 2					; discard leading '0x'
    mov eax,[byte esi-vref+serno2]
    call appnum						; prepend middle dword
    add ecx,byte 2
    mov eax,[byte esi-vref+signature]
    pop ioflagw						; rst formatting flags
    call appnum						; prepend m.s. dword
    mov edi,ecx						; string is long
    lea ecx,[edi-VNAME.NONE+CAP.SERNO]				; "psn "
    mov oflagb,o_ascil|o_lstg				; make long string available
    xor eax,eax
    inc eax						; numeric just notifies presence of long serno
    ret
    a
ssem:
    call sse
    test eax,eax
    jnz xmmx	;jz psn.r
    jr psn.r	;jr xmmx
;
;;  feature bits;##;;	:							;
fpu:	dec ebx	;0	;
vme:	dec ebx	;1	;
de:	dec ebx	;2	;
pse:	dec ebx	;3	;
tsc:	dec ebx	;4	;
msr:	dec ebx	;5	;
pae:	dec ebx	;6	;
mce:	dec ebx	;7	;
cx8:	dec ebx	;8	;
apic:	dec ebx	;9	; K5/m:0:global paging xtn
	dec ebx	;10	;
sep:	dec ebx	;11	;
mtrr:	dec ebx	;12	;
pge:	dec ebx	;13	: K5/m:0: 0 (reserved)
mca:	dec ebx	;14	;
cmov:	dec ebx	;15	;
pat:	dec ebx	;16	;
ps36:	dec ebx	;17	;
pn:	dec ebx	;18 pn	;I
	dec ebx	;19	;
	dec ebx	;20	;
	dec ebx	;21	;
xmmx:	dec ebx	;22	;A xtd
mmx:	dec ebx	;23	;
fxsr:	dec ebx	;24	;
simd:			;;I: intel "simd" is amd "sse" 		(re below)	;
sse:	dec ebx	;25	;IA
	dec ebx	;26	;
	dec ebx	;27	;
	dec ebx	;28	;
l64m:	dec ebx	;29	;A 'sledge hammer' 64bit "long mode"
x3dn:	dec ebx	;30	;A xtd
a3dn:	dec ebx	;31	;A xtd
    or oflagb,o_flag					; result is flag
    add ebx,byte(CAP.TOP-CAP.FBOT)/4-1			; requested type no.; 31 for "a3dn"
    mov eax,ebx
    inc eax						; return numeric 1..32
    lea ecx,[edi-VNAME.NONE+CAP.FPU+ebx*4]
    mov edx,[byte esi-vref+features]
    and edx,[edi-VNAME.NONE+asfbm]
    bt edx,ebx
    jc .p
    mov edx,[byte esi-vref+xfeat.d]
    and edx,[edi-VNAME.NONE+axfbm]
    bt edx,ebx
    jc .p
.u:
    xor eax,eax
.x:
    xor edx,edx		;;
    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]			; "unkn"
.p:
    and edx,[edi-VNAME.NONE+aspec]
    jnz .k						; bits 9,12,13 AMD special
    test ecx,ecx
.r:
    ret
    a
.k:;;					"gpe"		: K5 model 0 special	;
    push eax
    xor eax,eax
    call vamd
    pop eax
    jnz .r						;?; not AMD
    mov bx,0x0fff			 		; extract family/model/stepping
    and bx,[byte esi-vref+signature]
    mov dx,0000111111110000b
    and dx,bx
    cmp dx,0000010100100000b
    jnz .m						; other cpu, try mtrr query
    cmp al,14
    jz .u						; "unkn" bit 13
    cmp al,10
    jnz .r
    lea ecx,[edi-VNAME.NONE+CAP.GPE]			; "gpe" is bit 9
    ret
.m:;;					"mtrr"		: K6/Athlon special	;
    cmp al,13
    jnz .r
; Check support for AMD-K6 processor-style MTRRs:
    cmp bx,0x0588			 		;  C if CPU<K6-2/CXT
    sbb edx,edx 					; -1 if CPU<K6-2/CXT

    not edx			 			;  0 if CPU<K6-2/CXT
    cmp bx,0x600					;  C if CPU<Athlon
    sbb ebx,ebx						; -1 if CPU<AMD-K6
    and edx,ebx						; -1 if (CPU>=AMD-K6-2/CXT)&&(CPU<AMD Athlon)
    and eax,edx
    jz near .x
    ret    
;; restrict "simd" to intel models? -> uncomment		 (re above)	; :HIER:
;simd:							;I: intel simd is amd sse
;    call vintl
;    jz sse
;    lea ecx,[edi-VNAME.NONE+QUERY.UNKN]		; "unkn"
;    test ecx,ecx
;    ret

; - all current features list -
    a
feat:
    mov eax,[byte esi-vref+features]
    and eax,[edi-VNAME.NONE+asfbm]
    mov ebx,[byte esi-vref+xfeat.d]
    and ebx,[edi-VNAME.NONE+axfbm]
    or eax,ebx
    test iflagb,r_bin|r_oct|r_hex
    jnz n2s						;?; not bitwise
    push dword (CAP.TOP-CAP.FPU)/4-1
.l:
    mov ecx,[esp]
    add ecx,(CAP.FPU-QUERY.BOT)/4
    mov oflagb,0					; clear output flags
    call getcaps
    test iflagb,r_stgs					; suppress "unkn" string output
    jz .o
    test eax,eax
    jz .n
.o:
    test dword[esp],-1
    jng .n						; last bit by regular output
    call ret_val
.n:
    dec dword[esp]
    jns .l
    pop ebx
    xor ebx,ebx
    inc ebx
    ret

; i:	eax	number
; o:	eax=eax
;	ecx	numeric asciz string
;	edi	ecx
; c:	all but eax
    a
n2s:	;<;(unkny), (feat); if numeric			; build numeric returnstring from eax
    push eax
    call retnum
    mov edi,ecx						; -l is -s, num. stg
    mov oflagb,o_ascis|o_ascil|o_lstg
    pop eax
    test ecx,ecx
    ret

; - cache -
c1a:
;;    positions in (blockwise reverse) sync with names labels. don't change!	;  :OK:
; - AMD L1 - <cache_l1>
; - 5.eax - athlon+bit3 set
x1ie:	dec ebx	;L1.EITLB_ENTRIES 1   Extd (2/4 Mbyte) L1 Instruction Cache TLB Entries
x1ia:	dec ebx	;L1.EITLB_ASSOC	  2   Extd (2/4 Mbyte) L1 Instruction Cache TLB Ass
x1de:	dec ebx	;L1.EDTLB_ENTRIES 3   Extd (2/4 Mbyte) L1 Data Cache TLB Entries
x1da:	dec ebx	;L1.EDTLB_ASSOC	  4   Extd (2/4 Mbyte) L1 Data Cache TLB Ass (0xff = full)
; - 5.ebx - athlon+bit 3 unset - k5/k6
t1ie:	dec ebx	;L1.ITLB_ENTRIES  1   L1 Instruction Cache TLB Entries
t1ia:	dec ebx	;L1.ITLB_ASSOC	  2   L1 Instruction Cache TLB Associativity (0xff = full)
t1de:	dec ebx	;L1.DTLB_ENTRIES  3   L1 Data Cache TLB Entries
t1da:	dec ebx	;L1.DTLB_ASSOC	  4   L1 Data Cache TLB Associativity
; - 5.ecx
l1cz:	dec ebx	;L1.DCACHE_LSIZE  1   L1 Data Cache Line Size (bytes)
l1cl:	dec ebx	;L1.DCACHE_LINES  2   L1 Data Cache Lines
l1da:	dec ebx	;L1.DCACHE_ASSOC  3   L1 Data Cache Associativity (0xff = full)
l1ds:	dec ebx	;L1.DCACHE_SIZE	  4   L1 Data Cache Size (kbytes)
; - 5.edx	- ..15
l1iz:	dec ebx	;L1.ICACHE_LSIZE  1   L1 Instruction Cache Line Size (bytes)
l1il:	dec ebx	;L1.ICACHE_LINES  2   L1 Instruction Cache Lines
l1ia:	dec ebx	;L1.ICACHE_ASSOC  3   L1 Instruction Cache Associativity (0xff = full ass)
l1is:	dec ebx	;L1.ICACHE_SIZE	  4   L1 Instruction Cache Size (kbytes)
c1e:		;; ebx < 16 ;
    add ebx,byte(c1e-c1a)	; count begins from 0
    or oflagb,o_flag					; enable flag result
    xor eax,eax
    call vamd
    jnz c2e.x						;?; other vendor
    movzx eax,byte[byte esi-vref+cache_l1+ebx]
    test eax,eax
    jz c2e.q
    lea ecx,[edi-VNAME.NONE+L1.BOT+ebx*4]
.x:
    test ecx,ecx
.r:
    ret

; - AMD L2 - <cache_l2>
c2a:
; - 6.eax - athlon+bit 3 set	:HIER:	not checked!	:HIER:
; AAAA.EEEE,EEEE.EEEE:aaaa.eeee.eeee.eeee
x2da:	dec ebx	;A;L2.EDTLB_ASSOC	4.hn	Extd (2/4M) L2 Data Cache TLB Ass (0xf = full)
x2de:	dec ebx	;E;L2.EDTLB_ENTRIES	3+4l	Extd (2/4M) L2 Data Cache TLB Entries
x2ia:	dec ebx	;a;L2.EUTLB_ASSOC	2.hn	Extd (2/4M) L2 Instruction or Unif Cache TLB Ass
x2ie:	dec ebx	;e;L2.EUTLB_ENTRIES	1+2l	Extd (2/4M) L2 Inst or Unified Cache TLB Entries
; - 6.ebx - athlon+bit3 unset - k5/k6
; AAAA.EEEE,EEEE.EEEE:aaaa.eeee.eeee.eeee
t2da:	dec ebx	;A;L2.DTLB_ASSOC	4.hn	L2 Data Cache TLB Associativity
t2de:	dec ebx	;E;L2.DTLB_ENTRIES	3+4l	L2 Data Cache TLB Entries
t2ia:	dec ebx	;a;L2.UTLB_ASSOC	2.hn	L2 Instruction or Unified Cache TLB Ass (0xf = full)
t2ie:	dec ebx	;e;L2.UTLB_ENTRIES	1+2l	L2 Instruction or Unified Cache TLB Entries
; - 6.ecx - athlon+k6-3
; kkkk.kkkk,kkkk.kkkk:aaaa.tttt,bbbb.bbbb ;
l2us:	dec ebx	;k;L2.CACHE_SIZE	4+3	L2 Unified Cache Size (Kbytes)
l2ua:	dec ebx	;a;L2.CACHE_ASSOC	2.hn	L2 Unified Cache Associativity (0xf = full ass)
l2ul:	dec ebx	;t;L2.CACHE_LINES	2.ln	L2 Unified Cache Lines (lines per tag)
l2uz:	dec ebx	;b;L2.CACHE_LSIZE	1	L2 Unified Cache Line Size (bytes)
; - 6.edx - n.a. - reserved -
c2e:
    add ebx,byte(c2e-c2a)
    or oflagb,o_flag					; enable flag result
    xor eax,eax
    call vamd
    jnz .x						;?; not AMD
    lea ecx,[edi-VNAME.NONE+L1.BOT+(c1e-c1a)*4+ebx*4]
    mov edx,ebx
    and ebx,byte -2					; word disp
    movzx eax,word[byte esi-vref+cache_l2+ebx]
    jz .q						;?; descriptor not set
    sub edx,byte 8
    jz .r						;?; L2_CACHE_SIZE: l2[2]>>16)&0xffff
    jnc .l						;?; K6-3
    shr dl,1
    jnc .o
;L2_DTLB_ENTRIES: l2[0]>>16)&0xfff
;L2_UTLB_ENTRIES: l2[0]>> 0)&0xfff
;L2_EDTLB_ENTRIES:l2[1]>>16)&0xfff
;L2_EUTLB_ENTRIES:l2[1]>> 0)&0xfff
    and ah,byte 15
.r:
    test eax,eax
    jz .q						;?; descriptor not set
    test ecx,ecx
    ret
    a
.x:
    xor eax,eax
.q:
    lea ecx,[VENDOR.UNKN]				; parameter not applicable: "unkn"
    test ecx,ecx
    ret
.o:
;L2_DTLB_ASSOC:   l2[0]>>28)&0x0f
;L2_UTLB_ASSOC:   l2[0]>>12)&0x0f
;L2_EDTLB_ASSOC:  l2[1]>>28)&0x0f
;L2_EUTLB_ASSOC:  l2[1]>>12)&0x0f 
    and ax,byte 15
    jr .r
;- K6-3
    a
.l:
    xor ah,ah
    cmp edx,byte 3
;L2_CACHE_ASSOC:  l2[2]>>12)&0x0f
    jz .r
;L2_CACHE_LINES:  l2[2]>> 8)&0x0f

;L2_CACHE_LSIZE:  l2[2]>> 0)&0xff
    jr .o

; - intel cache -  2ndary levels in <xlevel>+	:HIER:	not checked!	:HIER:
it32:	dec ebx	;L2.IT32 	 Instr TLB, 4K Pages, 4-way set associative, 32 entries	;
it02:	dec ebx	;L2.IT04 	 Instr TLB, 4M Pages, fully associative, 2 entries	;
dt4k:	dec ebx	;L2.DT4K 	 Data TLB, 4K Pages, 4-way set associative, 64 entries	;
dt4m:	dec ebx	;L2.DT4N 	 Data TLB, 4M Pages, 4-way set associative, 8 entries	;
ic8: 	dec ebx	;L2.IC8: 	 Instr Cache, 8K, 4-way set associative, 32b line size	;
ic16:	dec ebx	;L2.IC16 	 Instr Cache, 16K, 4-way set associative, 32b line size	;
dc8: 	dec ebx	;L2.DC8: 	 Data Cache, 8K, 2-way set associative, 32b line size	;
dc16:	dec ebx	;L2.DC16 	 Data Cache, 16K, 2-way set associative, 32b line size	;
nl2: 	dec ebx	;L2.NL2: 	 No L2 Cache						;
uc1: 	dec ebx	;L2.UC1: 	 Unified, 32b cache line, 4-way set associative, 128K	;
uc2: 	dec ebx	;L2.UC2: 	 Unified, 32b cache line, 4-way set associative, 265K	;
uc5: 	dec ebx	;L2.UC5: 	 Unified, 32b cache line, 4-way set associative, 512K	;
uc1m:	dec ebx	;L2.UC1M Unified, 32b cache line, 4-way set associative, 1M	;
uc2m:	dec ebx	;L2.UC2M Unified, 32b cache line, 4-way set associative, 2M	;
.c:
    add ebx,byte(.c-it32+1)				; count begins at 1
    or oflagb,o_flag					; enable flag result
    xor eax,eax
    call vintl
    lea ecx,[VNAME]					; other vendor: "none"
    jnz .v
    call icregs
    lea ecx,[VENDOR.UNKN]				; parameter not applicable: "unkn"
    jz .r
    lea ecx,[L2.IT32-4+ebx*4]				; cache id-string
.r:
    test ecx,ecx
    ret
.v:
    xor eax,eax
    jr .r
; find cache config codebyte, eax/NZ if found
    a
icregs:
    mov dl,byte[icache-1+ebx]
    xor eax,eax
    push esi
    lea esi,[byte esi-vref+intel_20]
    mov al,byte[esi]					; no of levels to try
    test al,al
    jz .n						;?; no cache cfg.-> Z
    mov dh,al
.l:
    mov al,byte[esi]
    test al,al
    jz .n						;?; group not valid, ready
    mov ecx,3
    call rscan						; eax remainder
    jz .f
    call cscan						; ebx
    jz .f
    call cscan						; ecx
    jz .f
    call cscan						; edx
    jz .f
    dec dh
    jg .l						;?; more
    inc dh						; NZ
.f:
    setz al
    test al,al						; flag result
.n:
    pop esi
    movzx eax,al
    ret
; scan reg for codebyte, Z if found
    a
cscan:
    mov cl,4
rscan:
    lodsd
    test eax,eax
    js .x
.l:
    rol eax,8
    cmp al,dl
    loopnz .l
.x:
    ret

; complete cache config
    a
cach:
    lea edi,[edi-VNAME.NONE+acach]			; callcode table
    call qust.h						; call the cache part of '-s -?'
    and oflagb,~(o_quit|o_flag)				; invalidate quit command, inhibit flag results
    xor eax,eax						; mark ready
;    test esp,esp
    ret

; cpuid -? -?		returns all implemented querystring
; cpuid -n -?		can be used to finding the top index.
; cpuid -? query	returns the associated code number.
; cpuid -s -?		all currently applicable capability strings
    a
qust:
    mov bx,(QUERY_NOT_FOUND-QUERY.BOT)/4-1
    lea ecx,[edi-VNAME.NONE+VNAME]
    bts ioflagw,r.qst
    jc .q						;?: '-? -?'
    bt ioflagw,r.stg
    jc .s						;?: '-s -?'
    test iflagb,r_num					;;; '-n -?'
    jz .n						;?: '-? (query)'
.r:
    mov eax,ebx
.n:
    ret
    a
.q:							;;; "-? -?"
    lea ecx,[edi-VNAME.NONE+QUERY.FIRST]		; 1st stg
    push ecx
    lea edi,[byte(QUERY_NOT_FOUND-QUERY.FIRST)/4]	; count
    mov iflagb,r_stg					; force id-stg output
.l:
    mov ecx,[esp]					; ptr to stg
    push byte 4
    pop edx						; len
    add dword[esp],edx					; next stg
    cmp dword[ecx],'unkn'
    jz .u						;?; not a querystring
    call write						;>; send name
.u:
    dec edi
    jg .l
    pop ecx
.o:
    or oflagb,o_quit					; force no more output & exit -1
.x:
    xor eax,eax
    ret
    a
.e:
    add esp,byte 4
    jr .f
; currently available features
    a
.s:							;;; "-s -?"
    mov iflagb,r_stg					; clear other options
    lea edi,[edi-VNAME.NONE+avail]
.h:
    push edi
.f:
    mov oflagb,0					; clr output flags
    pop edi
    movzx ecx,byte[edi]					; table of disps
    inc edi						; next item
    test ecx,ecx
    jz .o						;?; ready
    test byte[edi],-1
    js .c						;?; sequence of queries
    push edi
    call getcaps
    jz .f						;?; no output, anyway
    mov edi,[esp]
    test eax,eax
    call ret_val					;>; write if eflags=NZ
    jr .f
; cache data    
    a
.c:
    movsx edx,byte[edi]
    inc edi
    push edi
    neg edx
    push ecx						; startindex*4
    push edx						; count
.g:
    pop edx
    dec edx
    js .e						;?; ready
    mov ecx,[esp]
    push edx
    add ecx,edx
    call getcaps
    jz .g
    test eax,eax
    call ret_val
    jr .g

; register values of a cpuid level, ret num. string: eax ebc ecx edx ;
    a
regs:
    xor ecx,ecx						; default is cpuid 0
    dec word[byte esi-vref+argc]			; next arg
    js .n						;?; no register argument
    pop edx						; ret addr
    xchg edx,[esp]					; edx:=argptr
    call st2ix						; fetch level code
; ecx:=num -> eax for cpuid call, 0x:standard level, 1x:intel cache, 8x:amd xtd
.n:
    push ioflagw
    test iflagb,r_digs
    jnz .o
    mov iflagb,r_hex					; default to sedecimal
.o:
    mov eax,ecx						; packed level code
    ror eax,4						; hi.nib:level lo.nib:flag - llll.0000....0000.ffff
    mov edi,eax
    and al,byte 8
    shl al,4
    bswap eax
    shr al,4						; cpuid code
    cmp di,1
    jnz .s						; standard/amd-xtd level
    bswap edi
    shr di,4						; 0000.0001....0000.cccc
    dec di
    js .c						; one step only
.i:
    mov eax,2						; intel cache
    cpu_id
    dec di
    jle .c						; requested level
    dec al
    jg .i						; continue if processor permits
    xor edi,edi
    pop ioflagw
    jr .x						; requested level not available
    a
.c:
    mov eax,2
.s:
    cpu_id						; fetch requested data
    push eax
    push ebx
    push ecx
    mov eax,edx						; edx
    call retnum
    pop eax						; ecx
    call .a
    pop eax						; ebx
    call .a
    pop eax
    call .a
    pop ioflagw
    or oflagb,o_ascil|o_lstg|o_ascis			; short & long string available & asciz
.x:
    mov eax,edi
.q:
    rol eax,4						; rst flag & num
    mov edi,ecx						; long:=short
    test esp,esp
    ret
    a
.a:
    dec ecx
    mov byte[ecx],' '
    xor ebx,ebx
    test iflagb,r_bin
    setnz bl
    sub ecx,ebx
    jmp appnum						; append numeric ascii to numbuf
    
; - emergency handler, index overflow -
    a
unkn:							; unknown request, ret nothing, or "0"
    lea edi,[quert]
    or oflagb,o_lstg|o_ascil
.s:
    lea ecx,[VENDOR.UNKN]
    xor eax,eax
    test ecx,ecx
    ret

;-
;- vamd
;-	specifically check for AMD
;- vintl
;-	specifically check for Intel
;- vcyrx
;-	specifically check for Cyrix
;- vcntr
;-	specifically check for Centaur
;- vumc
;-	specifically check for UMC
;- i:	eax=0
;- set eax=zero prior to processor specific call
;-
;- ckvend
;-	check vendor for cache type
;- i:	eax	vendor no.
;- o:	eflags	NZ if other vendor
;- c:	eax	vendor no.
;- p:	all other regs,oflags
;0: vunkn:
vumc:	inc eax	; 5
vcntr:	inc eax	; 4
vcyrx:	inc eax	; 3
vintl:	inc eax	; 2
vamd:	inc eax	; 1
ckvend:
    pushad
    push eax
    push ioflagw					; save i/o options
    xor ebx,ebx
    call vend
    pop ioflagw						; modifiers back
    pop ebx
    sub eax,ebx
    popad
    ret


;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --	;
;;	main program loop									;

%define __cmem	; clear unassigned memory
%undef  __cmem

    a
START:
%ifdef __cmem
    call cmem
%endif
    lea esi,[vref]					; ref to vari, never change!
    call detect_base					; detect cpuid presence, fetch data
    pop ecx						; argc
    dec ecx
    jz .h						;?; helptext if no args
    mov [byte esi-vref+argc],cx
    mov [byte esi-vref+argc+2],cx
    pop eax						; prog name
.n:
    xor ebx,ebx
    and oflagb,o_quit					; clear output flags, leave o_quit as is
    jne .r						;?; forced exit (vers, help)
    dec word[byte esi-vref+argc]
    js .q						;?; no (more) args
    pop edx						; next arg
    call st2ix						; find index from <[ea]> ptr to prmstg
.g:
    call getcaps					; evaluate
    call ret_val					; serve output
    jr .n						; try more
.h:
    push byte (OPTIONS.HELP-QUERY.BOT)/4		; no arguments default
    pop ecx
    jr .g
.r:
    inc ebx
.q:
    push ebx
    lea ecx,[eolt]
    call write.n					; send final <nl>
    pop ebx
    sys_exit

;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	auxilary, for data output								;

;-
;- ret_val, subroutine
;-	send cpuid reply to stdout
;- i:	eax	numeric val/flag
;-	ecx/edx	string, may also be a number, according to flags
;-	eflags	Z:ret immediately, no output
;- o:	eax	index
;-	ebx	flags
;-	ecx	stringptr
;-	edx	slen/-ve
    a
ret_val:
    jz write.r						;?; no output
    test iflagb,r_stgs
    jz rtnum						;?; not a string request
    call retstg
    jr write
    a
rtnum:
    call retnum
write:
    call .w
    test iflagb,r_bl
    jnz .r						;?; w.o. trailing blank
.b:
    lea ecx,[blank]
.n:
    push byte 1
    pop edx
.w:
    sys_write STDOUT
.r:
    ret

; -
;- retstg
;- i:	ecx	ptr to string
;-		asciz if o_ascis or if o_ascil and o_lstg
;-		or len=edx if o_lstg
;-		or len=4
;-	edx	long stg len if not o_ascil
;- o:	edx	strlen
;- c:	edx
    a
retstg:
    test iflagb,r_lstg
    jz .s						;?; not a long stg request
    test oflagb,o_lstg
    jz .s						;?; no long stg available, use short
    mov eax,edi
    test oflagb,o_dstg
    jnz retnum						; long from number edx
    mov ecx,edi						;   ptr, len is edx
    test oflagb,o_ascil
    jnz slen						;?; long is asciz
    ret
    a
.s:
    mov eax,ecx
    test oflagb,o_cstg
    jnz	retnum						;?; short from number ecx
    test oflagb,o_ascis
    jnz slen						;?; id is asciz	;
    push byte 4						;   id-string len
    pop edx
.r:
    ret
;-
;- slen
;-	find length of <nul>-terminated string
;- i:	ecx	ptr to string
;- o:	edx	strlen
;- c:	edx
    a
slen:
    mov edx,ecx
    dec edx
.l:
    inc edx
    test byte [edx],-1
    jnz .l
    sub edx,ecx
;.r:
    ret

;-
;- retnum
;-	according to <r_...>-flags send numeric string to stout
;-	number is eax
    a
retnum:
;	PRSTG':r#:'
    lea ecx,[byte esi-vref+numbuf.e-2]
    test iflagb,r_num
    jz .a						;?; fixed format
    test iflagb,r_digs
    jnz .a						;?; no forced flag if radix set
    test oflagb,o_flag
    jnz rt_hex						;?; flag result
.a:
    mov word[ecx],0					; prepare asciz
appnum:							;<; entry to prepend m.s. part of multiple regs num.
    test iflagb,r_hex
    jnz rt_hex						;?; sedecimal
    test iflagb,r_bin
    jnz rt_bin						;?; binary
    test iflagb,r_oct
    jnz rt_oct						;?; octal, else decimal
;-
;- rt_dec, subroutine
;-	ret signed decimal string
;- rt_bin, subroutine
;-	ret 32 digits 'b' postfixed binary string
;- rt_oct, subroutine
;-	ret 11 digits '0' prefixed octal string
;- rt_hex, subroutine
;-	ret 8 digits '0x' prefixed sedecimal string
;- i:	eax	num
;-	ecx	output buffer, endptr-1(!), trailing <nul> byte (numbuf.e-2)
;- o:	ecx	begin of buffer
;-	edx	string length
;- c:	eax,ecx,edx
    a
rt_dec:
    push byte 10
    pop edx						; get radix
    push ebx						; save reg
    xor ebx,ebx
    push ebx						; length count, +1 if binary
    test eax,eax
    jns .p
    inc byte[esp+3]					; signflag
    neg eax
.p:
    mov ebx,edx						; radix
; .-----------------.
.l:
    dec ecx						; numbuf-
    inc dword[esp]
    xor edx,edx						; hi dword for div
    div ebx
    or dl,'0'
    mov [ecx],dl
    test eax,eax					; ret S=0
    jnz .l
; `-----------------'
    test byte[esp+3],-1
    jz .s						;?; signed +ve
    mov bh,'-'
    mov byte[esp+3],0
.o:
    inc dword[esp]
    dec ecx
    mov byte[ecx],bh
.s:
    pop edx						; count
    pop ebx						; restore reg
    ret

; - 2^n radices, fixed format output
; i:	ecx	numbuf.e-2
;	eax	num
    a
rt_bin:
    mov bx,32<<8|1					; length | shift count
    mov byte[ecx],'b'
    jr rt_hex.n
    a
rt_oct:
    mov bx,11<<8|3
    jr rt_hex.n
    a
rt_hex:
    mov bx,8<<8|4
.n:
    push byte 1						; leading/trailing char
    test iflagb,r_num
    setnz dl
    jz .a						;?; free format not enabled
    test iflagb,r_digs
    jnz .a						;?; radix set, no flag
    test oflagb,o_flag
    jnz .f						;?; forced flag result
.a:
    dec dl
    and bh,dl
    movzx edx,bh
    mov edx,ecx						; buf (because <shift reg> only with cl)
    mov cx,bx						; length | shift count
    mov ebx,eax						; num
    xor eax,eax						; working reg (because daa is accu only)
    inc ch
.d:
    dec ch
.l:
    dec edx
    xor eax,eax	
    shrd eax,ebx,cl
    rol eax,cl
    add al,0x90
    daa
    adc al,0x40
    daa
    mov [edx],al
    inc dword[esp]					; count
    shr ebx,cl
    jnz .d
    dec ch
    jg .l						;?; more digits
    dec cl
    jz .r						;?; binary
    dec edx
    mov byte[edx],'0'
    cmp cl,2
    jz .r						;?; octal
    dec edx
    mov word[edx],'0x'
    inc dword[esp]					; adjust count
.r:
    mov ecx,edx						; buffer back
.p:
    pop edx						; count
    ret
    a
.f:							; ret flag w.o. affix
    test eax,eax
    setnz al
    or al,'0'
    mov [ecx],al
    pop edx
    ret
    

;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	auxilary, arguments fetching								;

;-
;- st2ix, subroutine
;-	get query index by string or number (dec, sedc, oct), find options
;- i:	edx	<[ea]> ptr to ptr to arg text
;- o	eax	arg with <nul>s replaced by <bl>s
;-	ecx	index
;-	edi	ptr to capstg
;- c:	eax, ecx, edi, eflags
    a
st2ix:
    push dword(QUERY.e-QUERY)>>2			; count
    mov eax,[edx]					; arg
    mov cl,' '
    bswap eax
    cmp ah,cl
    jnc .b
    mov ah,cl
    jr .c
    a
.b:
    cmp al,cl
    jnc .n
.c:
    mov al,cl
.n:
    bswap eax
    xor ecx,ecx						; clr numeric result
    cmp al,'0'
    jz st2h						; octal/sedec
    js .s						; try string
    cmp al,'9'
    jna st2n						; decimal
.s:
    lea edi,[QUERY]					; table
    mov ecx,[esp]
    repnz scasd
    inc ecx						; cap-stg index
    sub edi,byte 4					; pt to cap-stg
    sub [esp],ecx					; calc index
    pop ecx
    ret

; try acd number
; i:	eax	reg32, arg
;	[esp]	m32, maxarg
; o:	ecx	num
; c:	eax,ecx,edi
    a
st2n:	;<;st2ix.n
.l:
    sub al,'0'
    jc .n
    cmp al,9
    ja .n
    movzx edi,al
    lea ecx,[ecx+ecx*4]
    lea ecx,[edi+ecx*2]					; ecx 10 * al *
    shr eax,8						; NC flag "number"
    jnz .l						; ready
.n:
    pop edi						; maxcount
    cmp ecx,edi
    jng .r						; within range
    mov ecx,edi						; hi bound
.r:
    ret
; try sedec '0x'number, or '0'octal
; i:	[edx+2]	mem32, arg
;	[esp]	m32, maxarg
; o:	ecx	num
; c:	eax,ebx,ecx,edi
    a
st2h:	;<;st2ix
    mov ebx,ecx
    mov cx,8<<8|3
    cmp ah,'x'
    lea eax,[edx+1]
    jnz .o						;?; octal
    inc eax
    mov cx,16<<8|4
.o:
    mov eax,[eax]					; arg
.l:
    sub al,'0'
    jc .n
    cmp al,10
    jc .h
    sub al,7
    jc .n
    cmp al,ch
    jc .a
    sub al,('a'-'0'-17)
.h:
    cmp al,ch
    jnc .n
.a:
    shl ebx,cl
    or bl,al
    shr eax,8						; NC flag "number"
    jnz .l						; ready
.n:
    mov ecx,ebx
    jr st2n.n

;-
;- strncmp, subroutine
;-	exactly compare two strings, in either dir.
;-	internally limited to 64K of length, direction
;- i:	esi	ptr to source stg
;-	edi	ptr to  destn. stg
;-	ecx	max stringlen
;-	DF	set to required direction
;- o:	eflags	-ve:s>d, 0:s=d, +ve:s<d
;- c:	-/-
    a
strncmp:
    pushad
    jr vcmp.s
;-
;- vcmp, subroutine
;-	compare vendorstrings, w. memory increment
    a
vcmp:
    pushad
    cld
    lea esi,[byte esi-vref+idstr]			; long cpu name string
.s:
    xor eax,eax
    dec eax
    movzx ecx,cx					; clr stg-flag, limit to 64K
.c:
    test byte [esi],al
    jz .n
    test byte [edi],al
    jz .n
    cmpsb
    loopz .c
    popad
    ret
    a
.n:
    scasb
    popad
    ret

;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
;;	auxilary, other										;

%ifdef __cmem	;memory initially not clean?
    a
cmem:
    pushad
    xor eax,eax
    lea edi,[bot]
    lea ecx,[(top-bot)/4-1]
    rep stosd
    popad
    ret
%endif

;;	SIGILL signal handling for "ssef" if <orps xmmx..> tried and was "illegal..."		;

;:HIER: complicated, somehow, and most probably not as intended. but,it serves the purpose...:OK:

; clear "ssef" SIGILL redirection
; without explicitely resetting tos to <ebp> did restore the handler, as supposed, but 
; then returned to vioalating codeaddress to continuously procede with "illegal..."s.
; i:	ebx	return address
;	ebp	esp reference
;:HIER: => re linux../entry.S: ebp contains "current" (whatever that may be -> check!)
    a
;cannot be called, by signal structure, by no means, but should!!! - how?
asig:
    cmp ebp,esp
%ifdef __ssef
	jz near isegf.r
%endif
    jr usig.n
    a
usig:
    cmp ebp,esp
.n:
%ifdef __ssef
	jz near isegf.r
;	ret		; -> repetetive signals, 'ONESHOT' useless'?
%else
    jz isegf.r
%endif
    lea esp,[ebp]	;:HIER: (re below) drops 384 bytes
    push ebx		; ret to after violating code (passed from caller, where fault expected)
    pushad
    mov ebx,SIGILL	; restore signal
    lea ecx,[sr_r]	; sigaction new
    jr isegf.n

; initiate "ssef" SIGILL redirection
    a
isegf: ;<(ssef)
    pushad
    lea ecx,[sa_n]	;sigaction new
.n:
    lea edx,[sa_o]	;sigaction old, filled in by syscall/at ret from sighd)
    sys_sigaction SIGILL
    popad
    lea ebp,[esp+4]	;:HIER: that is it! but I doubt wheter the syscall was meant that way.
.r:
    ret
    a
p_regs:
	ret

    a
; - data structures - (re si_t, sa_o, sr_t in .bss, below)
sa_n: ;<(segf)
    dd usig		; handler (executes, endlessly! - "oneshot" useless <- how to advance over error addres???)
    dd NULL;1<<(SIGILL-1)	; sa_mask (0/SIGILL:endless, -1/~SIGILL:nothing)
; sa_oneshot: should restore default action immediately
; sa_siginfo: sa_siginfo routine of struct should be used - ...nebbich! nur rt-signale
    dd SA_ONESHOT
    dd NULL		; restorer (unused) - (nee...!)

sr_r: ;<(usig)
    dd SIG_DFL		; doesn't clear redirected handling, segfault if inverse
    dd NULL;1<<(SIGILL-1)	; re include/asm/signal.h
    dd SIG_DFL
    dd NULL

%ifdef __DBG
;;   -- displacements control (nasm doesn't warn!) --	;
disp1:
    db vref-numbuf					; below 129
    db intel_20-vref					; below 128
%endif
;-
;; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ;
    UDATASEG

    alignb 8

bot:
; -- disp <optf> to <vref> is 0x58 -- (max 0x80)
numbuf:		resd 9
.e:
optf:		resd 1		; current options
argc:		resd 1		; no. of args, counter
cpuname:	resd 1		; ptr to long name of cpu, flag if none or long serno

; Mist:
; <vref>-referenz faul, der ganze aufbau ist vermurkst, besser neu und systematisch :HIER: -> spter...
; data cache
, instruction specific
ibase:		resd 1		; flag whether cpuid possible and already determined
xindex:		resd 1		; no. of cpuid extended functions
clnum:		resd 1		; aux. counter
vendi:		resd 1
vendn:		resd 1		; found vendor number
xlevel:		resd 1		;I+A:0x80000000: eax:=no. of extended levels or, 2ndary Intel-levels
cputype:	resd 1		; 1:amd 2:intel 3:cyrix 4:centaur 5:umc 6:other
; - cpuid return regs eax..edx -
; data cache
, standard cpuid items
	cregb a_l1
signature:equ a_l1.a		;I+A:1: eax [family, model, stepping] = serno1, m.s.dw
features: equ a_l1.d		;I+A:1: edx standard feature bits
;-
	cregb a_l2		;I a: cpuid 3: edx, ecx
; -
	cregb a_l0
;idstr: <vref> esi ref. addr ; disp < 128 (32 dwords)
clevel:	equ a_l0.a		; X :eax cpuid 0: no. of standard query levels
idstr:	equ a_l0.b		; X :ebx cpuid 0: vendor identifier string, regs: a b d c
vref:
; data cache
, processor dependent, ext'd features
i_l0:	cregb xfeat		;  A: cpuid 0x80000001: extended features bits	;I: 0 (repeated)
i_l1:	cregb nstr		;  A: cpuid 0x80000002: processor name(3)	;I: 1 (repeated
i_l2:	cregb amd_x3		;  A: cpuid 0x80000003: processor name(2)	;I: 2 (cache
i_l3:	cregb amd_x4		;  A: cpuid 0x80000004: processor name(1)	;I: 3 (xtd signature
serno3:	equ amd_x4.c		;I  :		;I: ecx (lo 32bit of 96bit signature
serno2:	equ amd_x4.d		;I  :		;I: edx (mid 32bit of 96bit signature
	cregb cache_l1		;  A: cpuid 0x80000005: L1 TLB/cache information
	cregb cache_l2		;  A: cpuid 0x80000006: L2 TLB/cache information
; -- 0x70 last block at short disp to <vref> --
	cregb intel_20		;32	;I2 : cpuid 2 1st call, cache descriptors
	cregb intel_21			;I2: cpuid 2 consecutive calls, arbitrary max:=7
	cregb intel_22			;I2:
	cregb intel_23			;I2:
	cregb intel_24			;I2
	cregb intel_25			;I2
	cregb intel_26			;I2
	cregb intel_27			;I2
		resd 4*8	; spare

; - signals/handlers
si_t:		resd 32		; filled in by signal (oder auch nicht...)
so_t:		resd 32		; filled in by syscall
sa_o:		resd 32		; unused (?) <- falsely documented! -> receives the catched signal descr

top:

    END

;-
;-========================================================================;
;- cpuid.asm <eof>